/**********************************************************************
 * Utils.c: Miscellaneous support routines for xprzmodem.library;
 * Version 2.10, 12 February 1991, by Rick Huebner.
 * Released to the Public Domain; do as you like with this code.
 *
 * Version 2.50, 15 November 1991, by William M. Perkins.  Added code
 * to update_rate() function in utils.c to avoid the Guru # 80000005
 * you would have gotten if you had decided to adjust the system clock
 * back during an upload or download.
 *
 **********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include "types.h"
#include "xproto.h"
#include "zmodem.h"
#include "xprzmodem.h"

#define QVS(p1,p2) struct {short bra; char kets[sizeof(p2)]; } p1 = {0x3c3c,p2};

char _prog_name[] = "<QDOS XPR-ZMODEM v2.30>";

static QVS (_qcfx_, "QCFX>>01\x00\x09XPRZMODEM\x00\x00\x042.30\xFF\xFF");

#ifdef TST
#define DTEST(p1) else fputs(p1, stderr);
#else
#define DTEST(p1)
#endif

typedef struct SetupVars SETUP, *PSETUP;

/* Transfer options to use if XProtocolSetup not called */
static SETUP Default_Config =
{
#ifndef QDOS
    NULL, NULL, 0,
#endif
    {'C', 0},
    {'N', 0},
    {'1', '6', 0},
    {'0', 0},
    {'1', '0', 0},
    {'N', 0},
    {'N', 0},
    {'Y', 0},
    {'N', 0},
    {'Y', 0},
#ifdef _DCC
    {'T', ':', 0},
#else
    {'r','a','m','2','_',0}
#endif
};

static struct xpr_option op[] =
{
    {"ZModem options:", XPRO_HEADER, NULL, 0},
    {"Text mode (Y,N,?,C):", XPRO_STRING, NULL, 0},
    {"Overwrite mode (Y,N,R,S):", XPRO_STRING, NULL, 0},
    {"I/O buffer size (KB):", XPRO_LONG, NULL, 0},
    {"Frame size (bytes):", XPRO_LONG, NULL, 0},
    {"Error limit:", XPRO_LONG, NULL, 0},
    {"Auto-activate receiver:", XPRO_BOOLEAN, NULL, 0},
    {"Delete after sending:", XPRO_BOOLEAN, NULL, 0},
    {"Keep partial files:", XPRO_BOOLEAN, NULL, 0},
    {"Send full path:", XPRO_BOOLEAN, NULL, 0},
    {"Use received path:", XPRO_BOOLEAN, NULL, 0},
    {"Default receive path:", XPRO_STRING, NULL, 0}
};

#ifdef DEBUGLOG
UBYTE DebugName[40] = "XDebug_Log";
FILE *DebugLog = NULL;
#endif


/**********************************************************
 *	long XProtocolSetup(struct XPR_IO *xio)
 *
 * Called by comm program to set transfer options
 **********************************************************/
long XProtocolSetup (xio)
struct XPR_IO *xio;
{

    SETUP tempsv;
    PSETUP sv;
    char buf[256], *p;
    int i, len;

    /* Allocate memory for transfer options string */
    if ((sv = (PSETUP) xio->xpr_data) == NULL)
    {

	xio->xpr_data = (PSETUP) malloc (sizeof (SETUP));
	if ((sv = xio->xpr_data) == NULL)
	{
	    ioerr (xio, "Not enough memory");
	    return XPRS_FAILURE;
	}
	/* Start out with default options; merge user changes into defaults */
        
	movmem (&Default_Config, sv, sizeof (SETUP));
    }

    /* If options string passed by comm prog, use it; else prompt user */
    if (xio->xpr_filename)
    {
	strcpy (buf, xio->xpr_filename);
    }
    else
    {
	/* If xpr_options() implemented by comm program, use it */
	if (xio->xpr_extension >= 1 && xio->xpr_options)
	{
	    /*
	     * Let user edit temp copy of options so we can ignore invalid
	     * entries.  Have to init all this crud the hard way 'cause it's
	     * got to be on the stack in order to maintain reentrancy
	     */
	    movmem (sv, &tempsv, sizeof (SETUP));
	    op[1].xpro_value = (char *)tempsv.option_t;
	    op[1].xpro_length = sizeof (tempsv.option_t);
	    op[2].xpro_value = (char *)tempsv.option_o;
	    op[2].xpro_length = sizeof (tempsv.option_o);
	    op[3].xpro_value = (char *)tempsv.option_b;
	    op[3].xpro_length = sizeof (tempsv.option_b);
	    op[4].xpro_value = (char *)tempsv.option_f;
	    op[4].xpro_length = sizeof (tempsv.option_f);
	    op[5].xpro_value = (char *)tempsv.option_e;
	    op[5].xpro_length = sizeof (tempsv.option_e);
	    op[6].xpro_value = (char *)tempsv.option_a;
	    op[6].xpro_length = sizeof (tempsv.option_a);
	    op[7].xpro_value = (char *)tempsv.option_d;
	    op[7].xpro_length = sizeof (tempsv.option_d);
	    op[8].xpro_value = (char *)tempsv.option_k;
	    op[8].xpro_length = sizeof (tempsv.option_k);
	    op[9].xpro_value = (char *)tempsv.option_s;
	    op[9].xpro_length = sizeof (tempsv.option_s);
	    op[10].xpro_value = (char *)tempsv.option_r;
	    op[10].xpro_length = sizeof (tempsv.option_r);
	    op[11].xpro_value = (char *)tempsv.option_p;
	    op[11].xpro_length = sizeof (tempsv.option_p);

	    /* Convert Y/N used elsewhere into "yes"/"no" required by spec */
	    for (i = 6; i <= 10; ++i)
	    {
		strcpy (op[i].xpro_value, (op[i].xpro_value[0] == 'Y') ? "yes" : "no");
	    }

	    (*xio->xpr_options) (op, 12);

	    /* Convert "yes"/"no" or "on"/"off" into Y/N */
	    for (i = 6; i <= 10; ++i)
	    {
         if(op[i].xpro_value[0] == 'Y' ||
            op[i].xpro_value[0] == 'y' ||
            stricmp (op[i].xpro_value, "yes") == 0 ||
            stricmp (op[i].xpro_value, "on") == 0)
         {
                  op[i].xpro_value[0] = 'Y';
         }
         else
         {
                  op[i].xpro_value[0] = 'N';
         }

	    }
	    /* Convert xpr_options() results into parseable options string */
            sprintf(buf,"T%s,O%s,B%s,F%s,E%s,A%s,D%s,K%s,S%s,R%s,P%s",tempsv.option_t,
            tempsv.option_o,tempsv.option_b,tempsv.option_f,tempsv.option_e,
            tempsv.option_a,tempsv.option_d,tempsv.option_k,tempsv.option_s,
            tempsv.option_r,tempsv.option_p);
	}
	else
	{
	    /* Start buffer with current settings so user can see/edit them. */
            sprintf(buf,"T%s,O%s,B%s,F%s,E%s,A%s,D%s,K%s,S%s,R%s,P%s",sv->option_t,
            sv->option_o,sv->option_b,sv->option_f,sv->option_e,
            sv->option_a,sv->option_d,sv->option_k,sv->option_s,
            sv->option_r,sv->option_p);
	    if (xio->xpr_gets)
		(*xio->xpr_gets) ("ZModem options:", buf);
	}
    }
    /* Upshift options string for easier parsing */
    strupr (buf);

    /*
     * Merge new T(ext) option into current settings if given "TY" = Force
     * Text mode on, "TN" = Force Text mode off, "T?" = Use other end's text
     * mode suggestion (default to binary) "TC" = Ask Comm program for file
     * type
     */
    if ((p = (char *) find_option ((UBYTE *)buf, (UBYTE )'T')))
    {
	if (*p == 'Y' || *p == 'N' || *p == '?' || *p == 'C')
	    *sv->option_t = *p;
	else
	    ioerr (xio, "Invalid T flag ignored; should be Y, N, ?, or C");
    }
    DTEST("Error T\n")
    /*
     * Merge new O(verwrite) option into current settings if given "OY" =
     * Yes, delete old file and replace with new one, "ON" = No, prevent
     * overwrite by appending ".dup" to avoid name collision, "OR" = Resume
     * transfer at end of existing file, "OS" = Skip file if it already
     * exists; go on to next
     */
    if ((p = (char *) find_option ((UBYTE *)buf, (UBYTE)'O')))
    {
	if (*p == 'R' && !xio->xpr_finfo)
	    ioerr (xio, "Can't Resume; xpr_finfo() not supported");
	else if (*p == 'Y' || *p == 'N' || *p == 'R' || *p == 'S')
	    *sv->option_o = *p;
	else
	    ioerr (xio, "Invalid O flag ignored; should be Y, N, R, or S");
    }
    DTEST("Error O\n")

    /*
     * Merge new B(uffer) setting into current settings if given Size of file
     * I/O buffer in kilobytes
     */
    if ((p = (char *) find_option ((UBYTE *)buf, (UBYTE)'B')))
    {
	len = atoi ((char *)p);
	if (len < 1)
	    len = 1;
	sprintf ((char *) sv->option_b, "%ld", len);
    }
    DTEST("Error B\n")

    /*
     * Merge new F(ramelength) setting into other settings if given Number of
     * bytes we're willing to send or receive between ACKs. 0 = unlimited;
     * nonstop streaming data
     */
    if ((p = (char *) find_option ((UBYTE *)buf, (UBYTE)'F')))
    {
	len = atoi ((char *)p);
	if (len < 0)
	    len = 0;
	if (len > 0 && len < MINBLOCK)
	    len = MINBLOCK;
	sprintf ((char *)sv->option_f, "%ld", len);
    }
    DTEST("Error F\n")

    /*
     * Merge new E(rror limit) setting into other settings if given Number of
     * sequential errors which will cause an abort
     */
    if ((p = (char *) find_option ((UBYTE *)buf, (UBYTE)'E')))
    {
	len = atoi ((char *)p);
	if (len < 1)
	    len = 1;
	if (len > 32767)
	    len = 32767;
	sprintf ((char *)sv->option_e, "%ld", len);
    }
    DTEST("Error E\n")

    /*
     * Merge new A(uto-activate) setting into other settings if given "AY" =
     * Automatically call XProtocolReceive() if ZRQINIT string received "AN"
     * = Don't look for ZRQINIT; user will explicitly activate receive
     */
    if ((p = (char *) find_option ((UBYTE *)buf, (UBYTE)'A')))
    {
	if (*p == 'Y' || *p == 'N')
	    *sv->option_a = *p;
	else
	    ioerr (xio, "Invalid A flag ignored; should be Y or N");
    }
    DTEST("Error A\n")

    /*
     * Merge new D(elete after sending) setting into other options "DY" =
     * Delete files after successfully sending them "DN" = Don't delete files
     * after sending
     */
    if ((p = (char *) find_option ((UBYTE *)buf, (UBYTE)'D')))
    {
	if (*p == 'Y' && (xio->xpr_extension < 2 || !xio->xpr_unlink))
	    ioerr (xio, "Can't use DY; xpr_unlink() not supported");
	else if (*p == 'Y' || *p == 'N')
	    *sv->option_d = *p;
	else
	    ioerr (xio, "Invalid D flag ignored; should be Y or N");
    }
    DTEST("Error D\n")

    /*
     * Merge new K(eep partial files) setting into other options "KY" = Keep
     * partially-received file fragments to allow later resumption "KN" =
     * Delete partially-received file fragments
     */
    if ((p = (char *) find_option ((UBYTE *)buf, (UBYTE)'K')))
    {
	if (*p == 'N' && (xio->xpr_extension < 2 || !xio->xpr_unlink))
	    ioerr (xio, "Can't use KN; xpr_unlink() not supported");
	else if (*p == 'Y' || *p == 'N')
	    *sv->option_k = *p;
	else
	    ioerr (xio, "Invalid K flag ignored; should be Y or N");
    }
    DTEST("Error K\n")

    /*
     * Merge new S(end full path) setting into other options "SY" = Send full
     * filename including directory path to receiver "SN" = Send only simple
     * filename portion, not including directory path
     */
    if ((p = (char *) find_option ((UBYTE *)buf, (UBYTE)'S')))
    {
	if (*p == 'Y' || *p == 'N')
	    *sv->option_s = *p;
	else
	    ioerr (xio, "Invalid S flag ignored; should be Y or N");
    }
    DTEST("Error S\n")

    /*
     * Merge new R(eceive path) setting into other options "RY" = Use full
     * filename exactly as received; don't use P option path "RN" = Ignore
     * received directory path if any; use path from P option
     */
    if ((p = (char *) find_option ((UBYTE *)buf, (UBYTE)'R')))
    {
	if (*p == 'Y' || *p == 'N')
	    *sv->option_r = *p;
	else
	    ioerr (xio, "Invalid R flag ignored; should be Y or N");
    }
    DTEST("Error R\n")

    /*
     * Merge new P(ath) setting into other options "Pdir" = Receive files
     * into directory "dir" if RN selected "dir" can by any valid existing
     * directory, with or without trailing "/"
     */
    if ((p = (char *) find_option ((UBYTE *)buf, (UBYTE)'P')))
    {
	strcpy ((char *)sv->option_p, p);
	p = (char *)sv->option_p + strcspn ((char *)sv->option_p, " ,\t\r\n");
	*p = '\0';
    }
    DTEST("Error P\n")

    return (*sv->option_a == 'Y') ? XPRS_SUCCESS | XPRS_NORECREQ | XPRS_HOSTMON
	: XPRS_SUCCESS | XPRS_NORECREQ;
}				/* End of long XProtocolSetup() */

/**********************************************************
 *	long XProtocolCleanup(struct XPR_IO *xio)
 *
 * Called by comm program to give us a chance to clean
 * up before program ends
 **********************************************************/
long
 XProtocolCleanup (xio)
struct XPR_IO *xio;
{
    /* Release option memory, if any */
    if (xio->xpr_data)
    {
	free (xio->xpr_data);

	xio->xpr_data = NULL;
    }

    return XPRS_SUCCESS;
}				/* End of long XProtocolCleanup() */

/**********************************************************
 *	long XProtocolHostMon(struct XPR_IO *xio,
 *	   char *serbuff, long actual, long maxsize)
 *
 * Called by comm program upon our request (XPRS_HOSTMON)
 * to let us monitor the incoming data stream for our
 * receiver auto-activation string (ZRQINIT packet).
 * We only ask for this to be called if option AY is set.
 **********************************************************/
#ifndef QDOS
long
 XProtocolHostMon (xio, serbuff, actual, maxsize)
struct XPR_IO *xio;
char *serbuff;
long actual;
long maxsize;
{
    static UBYTE startrcv[] =
    {ZPAD, ZDLE, ZHEX, '0', '0', 0};
    struct SetupVars *sv;

    if (!(sv = (void *) xio->xpr_data))
	return actual;		/* XProtocolSetup() never called?! */

    if (!sv->matchptr)
	sv->matchptr = startrcv;

    /*
     * Scan through serbuff to see if we can match all bytes in the start
     * string in sequence.
     */
    for (sv->bufpos = (UBYTE *) serbuff; sv->bufpos < (UBYTE *) (serbuff + actual); ++sv->bufpos)
    {
	if (*sv->bufpos == *sv->matchptr)
	{			/* if data matches current position in match */
	    ++sv->matchptr;	/* string, increment match position */
	    if (!*sv->matchptr)
	    {			/* if at end of match string, it all matched */
		sv->buflen = actual - (sv->bufpos - (UBYTE *) serbuff);
		XProtocolReceive (xio);
		sv->matchptr = startrcv;
		actual = 0;
		break;
	    }
	}
	else if (sv->matchptr > startrcv)
	{			/* mismatch?  Reset to start of match string */
	    sv->matchptr = startrcv;
	    if (*sv->bufpos == *sv->matchptr)
		++sv->matchptr;
	}
    }

    sv->bufpos = NULL;
    return actual;
}

/**********************************************************
 *	long XProtocolUserMon(struct XPR_IO *xio,
 *	   char *serbuff, long actual, long maxsize)
 *
 * Called by comm program to let us monitor user's inputs;
 * we never ask for this to be called, but it's better to
 * recover gracefully than guru the machine.
 **********************************************************/
long
 XProtocolUserMon (xio, serbuff, actual, maxsize)
struct XPR_IO *xio;
char *serbuff;
long actual;
long maxsize;
{
    return actual;
}
#endif

/**********************************************************
 *	struct Vars *setup(struct XPR_IO *io)
 *
 * Perform setup and initializations common to both Send
 * and Receive routines
 **********************************************************/
struct Vars *setup (io)
struct XPR_IO *io;
{
    static int bauds[] =
    {75, 300, 600, 1200, 2400, 4800, 9600, 19200};
    struct SetupVars *sv;
    struct Vars *v;
    long origbuf;

#ifdef DEBUGLOG
    long i, *lng;
#endif

    /* Make sure comm program supports the required call-back functions */
    if (!io->xpr_update)
	return NULL;
    if (!io->xpr_fopen || !io->xpr_fclose || !io->xpr_fread
	|| !io->xpr_fwrite || !io->xpr_fseek || !io->xpr_sread
	|| !io->xpr_swrite)
    {
	ioerr (io, "Comm prog missing required function(s); see docs");
	return NULL;
    }

    /* Hook in default transfer options if XProtocolSetup wasn't called */
    if (!(sv = (void *) io->xpr_data))
    {
	io->xpr_data = (PSETUP) malloc ( sizeof (struct SetupVars));

	if (!(sv = (void *) io->xpr_data))
	{
	    ioerr (io, "Not enough memory");
	    return NULL;
	}
	movmem (&Default_Config, sv, sizeof (SETUP));
    }

    /* Allocate memory for our unshared variables, to provide reentrancy */
    if (!(v = (struct Vars *) malloc (sizeof (struct Vars))))
    {
	nomem:
	ioerr (io, "Not enough memory");
	return NULL;
    }
    setmem(v, sizeof(struct Vars), '\0');

#ifndef QDOS    
    v->Modemchar = v->Modembuf;
#endif
    /*
     * Allocate memory for our file I/O buffer; if we can't get as much as
     * requested, keep asking for less until we hit minimum before giving up
     */
#ifndef QDOS
    v->Filebufmax = origbuf = atoi ((char *) sv->option_b) * 1024;
    while (!((char *)v->Filebuf =  malloc ((unsigned) v->Filebufmax)))
    {
	if (v->Filebufmax > 1024)
	    v->Filebufmax -= 1024;
	else
	{
	    free (v);

	    goto nomem;
	}
    }

    /* If framelength was intended to match buffer size, stay in sync */
#endif
    v->Tframlen = atoi ((char *)sv->option_f);
#ifndef QDOS
    if (v->Tframlen && v->Tframlen == origbuf)
	v->Tframlen = v->Filebufmax;
#endif
    v->ErrorLimit = atoi ((char *)sv->option_e);

    /* Copy caller's io struct into our Vars for easier passing */
    movmem (io, &(v->io), sizeof (struct XPR_IO));

#ifdef DEBUGLOG
    if (!DebugLog)
    {
        char *p;
	if((p = getenv("XPR_DEBUG")))
	{
	    strcpy(DebugName, p);
	}
	DebugLog = fopen (DebugName, "a");
    }
    dlog (v, "XPR_IO struct:\n");
    for (i = 0, lng = (long *) io; i < sizeof (struct XPR_IO) / 4; ++i)

    {
	sprintf (v->Msgbuf, "  %08lx\n", *lng++);
	dlog (v, v->Msgbuf);
    }
#endif

    /* Get baud rate; set serial port mode if necessary (and possible) */
    if (v->io.xpr_setserial)
    {
	v->Oldstatus = (*v->io.xpr_setserial) (3L);
	v->Baud = bauds[7 - v->Oldstatus];
    }
    else
    {
	v->Baud = 2400;
    }

    return v;
}				/* End of struct Vars *setup() */

/**********************************************************
 *	void set_textmode(struct Vars *v)
 *
 * Set text/binary mode flags in accordance with T option
 * setting
 **********************************************************/
void
 set_textmode (v)
struct Vars *v;
{
    struct SetupVars *sv;
    long i;

    sv = (void *) v->io.xpr_data;
    switch (*sv->option_t)
    {
	case 'Y':		/* Force text mode on receive; suggest text
				 * mode on send */
	  TY:
	    v->Rxascii = TRUE;
	    v->Rxbinary = FALSE;
	    v->Lzconv = ZCNL;
	    break;
	case 'N':		/* Force binary mode on receive; suggest
				 * binary mode on send */
	  TN:
	    v->Rxascii = FALSE;
	    v->Rxbinary = TRUE;
	    v->Lzconv = ZCBIN;
	    break;
	case 'C':		/* Ask comm program for proper mode for this
				 * file */
	    if (v->io.xpr_finfo)
	    {
		i = (*v->io.xpr_finfo) (v->Filename, 2L);
		if (i == 1)	/* Comm program says use binary mode */
		    goto TN;
		if (i == 2)	/* Comm program says use text mode */
		    goto TY;
	    }
	    /* xpr_finfo() not provided (or failed); default to T? */
	case '?':
	    v->Rxascii = v->Rxbinary = FALSE;
	    v->Lzconv = 0;
	    break;
    }
}				/* End of void set_textmode() */

/**********************************************************
 *	UBYTE *find_option(UBYTE *buf, UBYTE option)
 *
 * Search for specified option setting in string
 **********************************************************/
UBYTE *find_option(UBYTE *buf, UBYTE option)
{
    while (*buf)
    {
        buf += strspn((char *)buf," ,\t\r\n");
        if (buf && *buf == option)
        {
            return ++buf;
        }
        buf += strcspn((char *)buf," ,\t\r\n");
    }
    return NULL;
}

/**********************************************************
 *	void canit(struct Vars *v)
 *
 * send cancel string to get the other end to shut up
 **********************************************************/
void
 canit (v)
struct Vars *v;
{
    static char canistr[] =
    {24, 24, 24, 24, 24, 24, 24, 24, 24, 24,
     8, 8, 8, 8, 8, 8, 8, 8, 8, 8, 0};

    zmputs (v, (UBYTE *)canistr);
}				/* End of void canit() */

/**********************************************************
 *	void zmputs(struct Vars *v, UBYTE *s)
 *
 * Send a string to the modem, with processing for \336 (sleep 1 sec)
 * and \335 (break signal, ignored since XPR spec doesn't support it)
 **********************************************************/
void
 zmputs (v, s)
struct Vars *v;
UBYTE *s;
{
    UBYTE c;

    while (*s)
    {
	switch (c = *s++)
	{
	    case '\336':
#ifndef _DCC
                mt_susjb(-1, 50, NULL);
#endif
	    case '\335':
		break;
	    default:
		sendline (v, c);
	}
    }
    sendbuf (v);
}				/* End of void zmputs() */

/**********************************************************
 *	void xsendline(struct Vars *v, UBYTE c)
 *
 * Write one character to the modem
 **********************************************************/
void
 xsendline (v, c)
struct Vars *v;
UBYTE c;
{
    v->Outbuf[v->Outbuflen++] = c;
    if (v->Outbuflen >= sizeof (v->Outbuf))
	sendbuf (v);
}				/* End of void xsendline() */

/**********************************************************
 *	void sendbuf(struct Vars *v)
 *
 * Send any data waiting in modem output buffer
 **********************************************************/
void
 sendbuf (v)
struct Vars *v;
{
    if (v->Outbuflen)
    {
	(*v->io.xpr_swrite) (v->Outbuf, (long) v->Outbuflen);
	v->Outbuflen = 0;
    }
}				/* End of void sendbuf() */

/**********************************************************
 *	short readock(struct Vars *v, short tenths)
 *
 * Get a byte from the modem;
 * return TIMEOUT if no read within timeout tenths of a
 * second, return RCDO if carrier lost or other fatal error
 * (sread returns -1).  Added in some buffering so we
 * wouldn't hammer the system with single-byte serial port
 * reads.  Also, the buffering makes char_avail() a lot
 * easier to implement.
 **********************************************************/
short
 readock (v, tenths)
struct Vars *v;
short tenths;
{
    short n;
    long t;

#ifndef QDOS

    /* If there's data waiting in our buffer, return next byte */
    if (v->Modemcount)
    {
      gotdata:
	v->Modemcount--;
	n = *v->Modemchar;
        v->Modemchar++;
	return n;
    }
    /*
     * Our buffer is empty; see if there's anything waiting in system buffer.
     * If the caller is in a hurry, don't wait around, but if it can spare a
     * half second, wait a bit and build up some input so we don't do as many
     * sread() calls.
     */
    t = tenths;

    v->Modemcount = (*v->io.xpr_sread) (v->Modembuf, 1, t);

    if (v->Modemcount < 1)		/* Nothing received in time */
	return TIMEOUT;
    v->Modemchar = v->Modembuf;	/* Reset buffer pointer to start of data */
    goto gotdata;
#else
    unsigned char c;

    t = tenths;

    n = (*v->io.xpr_sread) (&c, 1, t);
    if(n < 1)
    {
        return TIMEOUT;
    }
    else
    {
        return (short)c;
    }
#endif
}				/* End of short readock() */

/**********************************************************
 *	char char_avail(struct Vars *v)
 *
 * Check if there's anything available to read from the
 * modem
 **********************************************************/
char
 char_avail (v)
struct Vars *v;
{
#ifdef QDOS
    return CheckQueue();
#else
    if (v->Modemcount > 0)
	return TRUE;

    /* No data in our buffer; check system's input buffer */
    v->Modemcount = (*v->io.xpr_sread)	(v->Modembuf, 16, 0L);
    if (v->Modemcount < 1)	/* Nothing in system buffer either */
    {
	v->Modemcount = 0;
	return FALSE;
    }
    else
	/* System buffer had something waiting for us */
    {
	v->Modemchar = v->Modembuf;
	return TRUE;
    }
#endif    
}				/* End of char char_avail() */

/**********************************************************
 *	void update_rate(struct Vars *v)
 *
 * Update the elapsed time, expected total time, and
 * effective data transfer rate values for status display
 **********************************************************/
void
 update_rate (v)
struct Vars *v;
{
    ULONG sent, elapsed, expect;
    short hr, mins;
    int tv;

    /* Compute effective data rate so far in characters per second */
  
    sent = v->xpru.xpru_bytes - v->Strtpos;
    time (&tv);
    elapsed = tv - v->Starttime;

    /*
     * If we haven't transferred anything yet (just starting), make
     * reasonable guess (95% throughput); otherwise, compute actual effective
     * transfer rate
     */
         
    v->xpru.xpru_datarate = (sent > 0 && elapsed > 0) ? (sent / elapsed)
	: (v->Baud * 85 / 1000);

    /* Compute expected total transfer time based on data rate so far */
    if (v->xpru.xpru_filesize <= 0 || v->xpru.xpru_datarate == 0)
	expect = 0;		/* Don't know filesize; display time=0 */
    else 
	expect = (v->xpru.xpru_filesize - v->Strtpos) / v->xpru.xpru_datarate;

    if(expect < 0)
    {
        expect = 0;
    }
    if(expect > 86400)
    {
        expect = 86400;
    }
    hr = expect / 3600;		/* How many whole hours */
    expect -= hr * 3600;	/* Remainder not counting hours */
    mins = expect / 60;		/* How many whole minutes */
    expect -= mins * 60;	/* Remaining seconds */
    sprintf ((char *)v->Msgbuf, "%02ld:%02ld:%02ld", (long) hr, (long) mins, expect);
    v->xpru.xpru_expecttime = (char *) v->Msgbuf;

    /* Compute elapsed time for this transfer so far */
    if(elapsed < 0)
    {
        elapsed = 0;
    }
    if(elapsed > 86400)
    {
        elapsed = 86400;
    }
    hr = elapsed / 3600;
    elapsed -= hr * 3600;
    mins = elapsed / 60;
    elapsed -= mins * 60;
    sprintf ((char *) v->Msgbuf + 20, "%02ld:%02ld:%02ld", (long) hr, (long) mins,
	     elapsed);
    v->xpru.xpru_elapsedtime = (char *) v->Msgbuf + 20;
}				/* End of void update_rate() */

#ifndef QDOS
/**********************************************************
 *	long bfopen(struct Vars *v, UBYTE *mode)
 *
 * Buffered file I/O fopen() interface routine
 **********************************************************/
long
 bfopen (v, mode)
struct Vars *v;
UBYTE *mode;
{
    /* Initialize file-handling variables */
    v->Filebufpos = v->Filebuflen = v->Filebufcnt = 0;
    v->Fileflush = FALSE;
    v->Filebufptr = v->Filebuf;
    /* Open the file */
#ifdef DEBUGLOG
    sprintf (v->Msgbuf, "bfopen: %s %s\n", v->Filename, mode);
    dlog (v, v->Msgbuf);
#endif
    return (*v->io.xpr_fopen) (v->Filename, mode);
}				/* End of long bfopen() */

/**********************************************************
 *	void bfclose(struct Vars *v)
 *
 * Buffered file I/O fclose() interface routine
 **********************************************************/
void
 bfclose (v)
struct Vars *v;
{
    if (v->File)
    {
	/* If bfwrite() left data in buffer, flush it out before closing */
	if (v->Fileflush)
	    (*v->io.xpr_fwrite) (v->Filebuf, 1L, v->Filebufcnt, v->File);
	/* Close the file */
	(*v->io.xpr_fclose) (v->File);
	v->File = NULL;
    }
}				/* End of void bfclose() */

/**********************************************************
 *	void bfseek(struct Vars *v, long pos)
 *
 * Buffered file I/O fseek() interface routine
 **********************************************************/
void
 bfseek (v, pos)
struct Vars *v;
long pos;
{
    long offset;

    /*
     * If new file position is within currently buffered section, reset
     * pointers
     */
    if (pos >= v->Filebufpos && pos < v->Filebufpos + v->Filebuflen)
    {
	offset = pos - v->Filebufpos;
	v->Filebufptr = v->Filebuf + offset;
	v->Filebufcnt = v->Filebuflen - offset;
	/*
	 * Otherwise, fseek() file & discard buffer contents to force new
	 * read
	 */
    }
    else
    {
	(*v->io.xpr_fseek) (v->File, pos, 0L);
	v->Filebuflen = v->Filebufcnt = 0;
	v->Filebufpos = pos;
    }
}				/* End of void bfseek() */

/**********************************************************
 *	long bfread(struct Vars *v, UBYTE *buf, long length)
 *
 * Buffered file I/O fread() interface routine
 **********************************************************/
long
 bfread (v, buf, length)
struct Vars *v;
UBYTE *buf;
long length;
{
    long count, total = 0;

    /* Keep going until entire request completed */
    while (length > 0)
    {
	/* Copy as much of the request as possible from the buffer */
	count = (length <= v->Filebufcnt) ? length : v->Filebufcnt;
	CopyMem (v->Filebufptr, buf, count);
#ifdef DEBUGLOG
	sprintf (v->Msgbuf, "bfread got %ld bytes from buffer\n", count);
	dlog (v, v->Msgbuf);
#endif
	buf += count;
	total += count;
	length -= count;
	v->Filebufptr += count;
	v->Filebufcnt -= count;

	/* If we've emptied the buffer, read next buffer's worth */
	if (!v->Filebufcnt)
	{
	    v->Filebufpos += v->Filebuflen;
	    v->Filebufptr = v->Filebuf;
	    v->Filebufcnt = v->Filebuflen
		= (*v->io.xpr_fread) (v->Filebuf, 1L, v->Filebufmax, v->File);
#ifdef DEBUGLOG
	    sprintf (v->Msgbuf, "bfread read %ld bytes\n", v->Filebufcnt);
	    dlog (v, v->Msgbuf);
#endif
	    /* If we hit the EOF, return with however much we read so far */
	    if (!v->Filebufcnt)
		break;
	}
    }
    return total;
}				/* End of long bfread() */

/**********************************************************
 *	long bfwrite(struct Vars *v, UBYTE *buf, long length)
 *
 * Buffered file I/O fwrite() interface routine
 **********************************************************/
long
 bfwrite (v, buf, length)
struct Vars *v;
UBYTE *buf;
long length;
{
    long count, total = 0;

    /* Keep going until entire request completed */
    while (length > 0)
    {
	/* Copy as much as will fit into the buffer */
	count = v->Filebufmax - v->Filebufcnt;
	if (length < count)
	    count = length;
	CopyMem (buf, v->Filebufptr, count);
#ifdef DEBUGLOG
	sprintf (v->Msgbuf, "bfwrite buffered %ld bytes\n", count);
	dlog (v, v->Msgbuf);
#endif
	buf += count;
	total += count;
	length -= count;
	v->Filebufptr += count;
	v->Filebufcnt += count;
	v->Fileflush = TRUE;

	/* If we've filled the buffer, write it out */
	if (v->Filebufcnt == v->Filebufmax)
	{
	    count = (*v->io.xpr_fwrite) (v->Filebuf, 1L, v->Filebufcnt, v->File);
#ifdef DEBUGLOG
	    sprintf (v->Msgbuf, "bfwrite wrote %ld bytes\n", count);
	    dlog (v, v->Msgbuf);
#endif
	    if (count < v->Filebufcnt)
		return -1;
	    v->Filebufptr = v->Filebuf;
	    v->Filebufcnt = 0;
	    v->Fileflush = FALSE;
	}
    }
    return total;
}				/* End of long bfwrite() */
#endif

/**********************************************************
 *	void ioerr(struct XPR_IO *io, char *msg)
 *
 * Have the comm program display an error message for us,
 * using a temporary XPR_UPDATE structure; used to display
 * errors before Vars gets allocated
 **********************************************************/
void
 ioerr (io, msg)
struct XPR_IO *io;
char *msg;
{
    struct XPR_UPDATE xpru;

    if (io->xpr_update && msg)
    {
       	xpru.xpru_updatemask = XPRU_ERRORMSG;
	xpru.xpru_errormsg = msg;
	(*io->xpr_update) (&xpru);
    }
}				/* End of void ioerr() */

/**********************************************************
 *	void upderr(struct Vars *v, char *msg)
 *
 * Have the comm program display an error message for us, using the
 * normal XPR_IO structure allocated in Vars
 **********************************************************/
void
 upderr (v, msg)
struct Vars *v;
char *msg;
{
    v->xpru.xpru_updatemask = XPRU_ERRORMSG;
    v->xpru.xpru_errormsg = msg;
    if (msg == (char *) v->Msgbuf)	/* Ensure message length < 50 */
	msg[48] = '\0';
    (*v->io.xpr_update) (&v->xpru);
#ifdef DEBUGLOG
    dlog (v, msg);
    dlog (v, "\n");
#endif
}				/* End of void upderr() */

/**********************************************************
 *	void updmsg(struct Vars *v,char *msg)
 *
 * Have the comm program display a normal message for us
 **********************************************************/
void
 updmsg (v, msg)
struct Vars *v;
char *msg;
{
    if(msg)
    {
    v->xpru.xpru_updatemask = XPRU_MSG;
    v->xpru.xpru_msg = msg;
    if (msg == (char *) v->Msgbuf)	/* Ensure message length < 50 */
	msg[48] = '\0';
    (*v->io.xpr_update) (&v->xpru);
#ifdef DEBUGLOG
    dlog (v, msg);
    dlog (v, "\n");
#endif
}
}				/* End of void updmsg() */

/**********************************************************
 *	long getfree(void)
 *
 * Figure out how many bytes are free on the drive we're uploading to.
 * Stubbed out for now; not supported by XPR spec.
 **********************************************************/
long
 getfree ()
{
    return 0x7FFFFFFF;
}				/* End of long getfree() */

/**********************************************************
 *	char exist(struct Vars *v)
 *
 * Check whether file already exists; used to detect
 * potential overwrites
 **********************************************************/
char
 exist (v)
struct Vars *v;
{
    long file;

    file = (*v->io.xpr_fopen) (v->Filename, "r");
    if (file)
    {
	(*v->io.xpr_fclose) (file);
	return TRUE;
    }
    else
	return FALSE;

}				/* End of char exist() */


#ifdef DEBUGLOG
/**********************************************************
 *	void dlog(struct Vars *v, UBYTE *s)
 *
 * Write a message to the debug log
 **********************************************************/
void
 dlog (v, s)
struct Vars *v;
UBYTE *s;
{
    /* Open the debug log if it isn't already open */
    if (!DebugLog)
    {
	DebugLog = fopen (DebugName, "a");
    }
    fwrite (s, 1L, (long) strlen (s), DebugLog);
    /*
     * Close file to flush output buffer; comment these two lines out if you
     * aren't crashing your system and don't mind waiting until the transfer
     * finishes to look at your log file. 
     */
     fclose(DebugLog);
     DebugLog = NULL;
}				/* End of void dlog() */

#endif
/* End of Utils.c source */
